/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \file
  * This code solves the steady flow inside an aneurysm. It introduces several
  * new concepts like Guo off lattice boundary conditions, reading of
  * surface geometry STL files, smooth grid refinement and voxelization.
  * Make sure to unpack the file aneurysm.stl.tgz before running the
  * simulation.
  **/

#include "palabos3D.h"
#include "palabos3D.hh"


#include "../functionalsAndDescriptors5/thrombosisFunctional3D.h"
#include "../functionalsAndDescriptors5/thrombosisFunctional3D.hh"

#include "../functionalsAndDescriptors5/thrombosisLattices3D.h"
#include "../functionalsAndDescriptors5/thrombosisLattices3D.hh"

#include "../functionalsAndDescriptors5/initializationFunctional3D.h"
#include "../functionalsAndDescriptors5/initializationFunctional3D.hh"

#include "../functionalsAndDescriptors5/normalAndTangent3D.h"



using namespace plb;
using namespace std;

typedef double T;
typedef Array<T,3> Velocity;
#define DESCRIPTOR descriptors::FreeSurfaceSedimentationD3Q19Descriptor
#define DESCRIPTOR2 descriptors::AdvectionDiffusionD3Q7Descriptor

string outputDir = "./tmp/";


plint extraLayer      = 0;  // Make the bounding box larger; for visualization purposes
                            //   only. For the simulation, it is OK to have extraLayer=0.
const plint blockSize = 25; // Zero means: no sparse representation.
const plint envelopeWidth = 1;  // For standard BGK dynamics.
const plint extendedEnvelopeWidth = 2;  // Because the Guo off lattice boundary condition
                                        //   needs 2-cell neighbor access.

const plint margin = 4 ; // Extra margin of allocated cells around the obstacle.
const plint borderWidth = 1; // Because the Guo boundary condition acts in a one-cell layer.

// const T iniRhoFluid = T(1);
const T rhoEmpty    = T(1);
Array<T,3> forceOrientation(T(),T(),(T)1); //gravity along z-axis
Array<T,3> externalForce;




bool onCluster;
T dt;

//below, "i" stands for inlet and "o" stands for outlet:
T Xi0; // inlet is 1 cell thick, so no need for Xi0 and Xi1
T Xi1;
T Yi0;
T Yi1;
T Zi0 ; // Zi1 is determined by Zi0 and waterLevel
T Zi1 ;

T Xo0;
T Xo1; // Xo0 = margin
T Yo0;
T Yo1;
T Zo0;
T Zo1;


T outletVolPhys = 0.;
T inletSurfPhys = 0.;
T maxVelocityInletPhys = 0.;
T maxVelocityInletLB 	= 0.;
T cSmago		= 0.;
T kappa			= 0.;


T particleNumber = 3500; // number of injected particles
T iniParts = 0.;
T stressThreshold = (T)1.4e-3; // upper bound in order to thrombose (above erosion is active)
T criticalTime = (T)10; // minimum residence time in order to thrombose
T criticalNumber = (T)5000; // critical number of particles to have in order to have a possible thrombosis


T settlingVelocityLatticeUnits = 0.;
plint referenceResolution  = 0.;
T fluidDensity             = 0.;
T userDefinedInletDiameter = 0.;

T gPhys = 0.;
T nuRU = 0.;

plint referenceDirection = 0;

T simTime = 0;

TriangleSet<T>* outerTriangleSet = 0;
TriangleSet<T>* innerTriangleSet = 0;
TriangleSet<T>* visuTriangleSet = 0;
T currentTime = 0;

plint beginParticleDynamics = 1000; // 5000
plint callDepositionErosion = 300;
plint removeMassIter;
plint minStatIter;

bool useParticles;
bool vtk1;
bool vtk2;
bool interfaceSTL;
bool distanceSTL;
plint dataWrite = 200;

T waterLevelFactor;

// global pourrit variables
Box3D inletDomain;
Box3D outletDomain;
Box3D outletDomainParticles;
Box3D inletDomainParticles;

Box3D depositionErosionZone; // the region of the river where sedimentation and erosion are applied !

//Returns box coordinates in the stl syst of coord. Input coordinates : relative coords (each in [0,1])
Array<plint, 3> getStlCoordFromReal(T x, T y, T z, Box3D& box)
{
    plint xCoord = box.getNx() * x;
    plint yCoord = box.getNy() * y;
    plint zCoord = box.getNz() * z;
    return Array<plint,3> (xCoord, yCoord, zCoord);
}

//Opposite thing of getStlCoordFromReal
Array<T, 3> getRealCoordFromStl(plint x, plint y, plint z, Box3D& box)
{
    T xCoord = (T) x / box.getNx();
    T yCoord = (T) y / box.getNy();
    T zCoord = (T) z / box.getNz();
    return Array<T, 3>(xCoord, yCoord, zCoord);
}

void writeInterfaceSTL( MultiScalarField3D<T>& volumeFraction, plint iT)
{

    if(interfaceSTL) //interface stl
    {
        pcout << "writeInterfaceSTL at iT = " << iT << endl;
        std::vector<T> isoLevels;
        isoLevels.push_back((T) 0.5);
        typedef TriangleSet<T>::Triangle Triangle;
        std::vector<Triangle> triangles;

        std::auto_ptr<MultiScalarField3D<T> > smoothVolumeFraction (
                     lbmSmoothen<T,DESCRIPTOR> (
                       *lbmSmoothen<T,DESCRIPTOR>(volumeFraction, volumeFraction.getBoundingBox()),
                         volumeFraction.getBoundingBox()) );

        Box3D interfaceBox = smoothVolumeFraction->getBoundingBox();
        interfaceBox.z1 -= 10;
        isoSurfaceMarchingCube(triangles, *smoothVolumeFraction, isoLevels, smoothVolumeFraction->getBoundingBox());
        const string path = outputDir + createFileName("interface", iT, 6) + ".stl";
        pcout << "Path = " << path << std::endl;
        TriangleSet<T>(triangles).writeBinarySTL(path);
    }
    else{
        pcout << "Couldn't write STL of interface at iT = " << iT << endl;
    }
}

void writeUVtk(MultiBlockLattice3D<T,DESCRIPTOR>& lattice,
               Box3D const& vtkDomain,
               std::string fname,
               Array<T,3> location,
               T dx,
               T dt,
               plint iter)
{
    pcout << "Writing U vtk" << std::endl;
    VtkImageOutput3D<T> uVtk(outputDir + createFileName("uVtk_" + fname, iter, 8), dx, location);
    uVtk.writeData<3,float>(*computeVelocity(lattice, vtkDomain), "latticeVel", dx/dt);
}

void writeVtkPlus ( MultiBlockLattice3D<T,DESCRIPTOR>& lattice,
                    MultiBlockLattice3D<T,DESCRIPTOR2>* particles,
                    MultiScalarField3D<T>& volumeFraction,
                    MultiScalarField3D<int>& flagMatrix,
                    MultiScalarField3D<T>& flagMatrixPlot,
                    MultiScalarField3D<T>* particlesDynamicsPlot,
                    MultiScalarField3D<T>* erosionDepositionPlot,
                    MultiScalarField3D<T>* distanceFunction,
                    MultiScalarField3D<T>* wss,
                    MultiScalarField3D<T>* temporalMeanWss,
                    Box3D const& vtkDomain,
                    std::string fname,
                    Array<T,3> location,
                    T dx,
                    T dt,
                    plint iter )
{

    pcout << "Write vtk plus" << std::endl;
    VtkImageOutput3D<double> dynVtk(outputDir + createFileName("dynVtk_" + fname, iter, 8), dx, location);
    dynVtk.writeData<float>(*extractSubDomain(*particlesDynamicsPlot, vtkDomain), "particlesDynamics");


    if(vtk1)
    {

        VtkImageOutput3D<double> vtkOut(createFileName("vtk_" + fname + "_", iter, 8), dx, location);
        vtkOut.writeData<float>(*computeVelocityNorm(lattice, vtkDomain), "latticeVelNorm", (T)1);
        vtkOut.writeData<float>(*computeDensity(*particles, vtkDomain), "particlesDensity", T(1));
        vtkOut.writeData<float>(*extractSubDomain(volumeFraction, vtkDomain), "vof");
        vtkOut.writeData<float>(*extractSubDomain(flagMatrixPlot, vtkDomain), "flagExternal");
        vtkOut.writeData<float>(*extractSubDomain(*wss, vtkDomain), "wss");

        if (useParticles)
        {
            vtkOut.writeData<float>(*extractSubDomain(*temporalMeanWss,vtkDomain), "temporalMeanWss");
            vtkOut.writeData<float>(*extractSubDomain(*erosionDepositionPlot,vtkDomain), "erosionDepositionFlag");
            //vtkOut.writeData<float>(*extractSubDomain(*particlesDynamicsPlot,vtkDomain), "particlesDynamics");
            vtkOut.writeData<float>(*copyConvert<int,double>(*extractSubDomain(flagMatrix, vtkDomain)), "bounceBack", 1.);
            vtkOut.writeData<float>(*extractSubDomain(*distanceFunction,vtkDomain), "distance", (T)1);
        }
    }

    if (distanceSTL)
    {
        pcout << "Write distance stl" << std::endl;
        Box3D eroBB = distanceFunction->getBoundingBox();
        //volFracBB.x1 = volFracBB.x1;
        std::vector<T> isoLevels;
        isoLevels.push_back((T) 0.01);
        typedef TriangleSet<T>::Triangle Triangle;
        std::vector<Triangle> triangles;
        isoSurfaceMarchingCube(triangles, *distanceFunction, isoLevels, eroBB);

        TriangleSet<T> triangleSet(triangles);
        triangleSet.scale(dx);
        triangleSet.translate(location);

        DEFscaledMesh<T> defMesh((TriangleSet<T>(triangles)));
        //defMesh.getMesh().smooth();
        defMesh.getMesh().smooth(3);
        //triangleSet.writeAsciiSTL(myPath + createFileName("distanceFunction", iter, 6) + ".stl");
        triangleSet.writeBinarySTL(outputDir + createFileName("distanceFunction", iter, 6) + ".stl");
    }
}


void writeVtkPlus2 (MultiBlockLattice3D<T,DESCRIPTOR>& lattice,
                    MultiBlockLattice3D<T,DESCRIPTOR2>* particles,
                    MultiScalarField3D<T>& volumeFraction,
                    MultiScalarField3D<int>& flagMatrix,
                    MultiScalarField3D<T>& flagMatrixPlot,
                    MultiScalarField3D<T>* particlesDynamicsPlot,
                    MultiScalarField3D<T>* erosionDepositionPlot,
                    MultiScalarField3D<T>* distanceFunction,
                    MultiScalarField3D<T>* wss,
                    MultiScalarField3D<T>* temporalMeanWss,
                    Box3D const& vtkDomain,  std::string fname,
                    Array<T,3> location, T dx, T dt, plint iter )
{

    pcout << "Write vtk plus 2" << std::endl;
    VtkImageOutput3D<double> vtkOut(createFileName("vtk_"+fname+"_", iter, 8), dx,location);
    vtkOut.writeData<float>(*computeVelocityNorm(lattice,vtkDomain), "uNorm", (T)1);
    vtkOut.writeData<float>(*extractSubDomain(volumeFraction,vtkDomain), "vol_frac");

    if (useParticles)
    {
        vtkOut.writeData<float>(*extractSubDomain(*temporalMeanWss,vtkDomain), "temporalMeanWss");
        vtkOut.writeData<float>(*extractSubDomain(*erosionDepositionPlot,vtkDomain), "erosionDepositionFlag");
        vtkOut.writeData<float>(*extractSubDomain(*distanceFunction,vtkDomain), "distance", (T)1);
    }

}


void writeVTKwholeLattice ( MultiBlockLattice3D<T,DESCRIPTOR>& lattice,
                            MultiBlockLattice3D<T,DESCRIPTOR2>* particles,
                            MultiScalarField3D<T>& volumeFraction,
                            MultiScalarField3D<int>& flagMatrix,
                            MultiScalarField3D<T>& flagMatrixPlot,
                            MultiScalarField3D<T>* particlesDynamicsPlot,
                            MultiScalarField3D<T>* erosionDepositionPlot,
                            MultiScalarField3D<T>* distanceFunction,
                            MultiScalarField3D<T>* wss,
                            MultiScalarField3D<T>* temporalMeanWss,
                            Array<T,3> location,
                            T dx,
                            T dt,
                            plint iter)
{
    if(useParticles)
    {

        pcout << "      Entering write whole" << std::endl;
        Array<T,3> maxX_lowestY(2107, 549, 8); // extreme right
        Array<T,3> maxY_lowestX(1857, 720, 32); // extreme left
        // Array bigWholeParty for velocity
        Array<T,3> maxX_lowestY_lu((maxX_lowestY-location)/dx); // to lattice Units
        Array<T,3> maxY_lowestX_lu((maxY_lowestX-location)/dx);  // to lattice Units


        Box3D yz_vtkWhole ( util::roundToInt(maxY_lowestX_lu[0]), util::roundToInt(maxX_lowestY_lu[0]),
                            util::roundToInt(maxX_lowestY_lu[1]), util::roundToInt(maxY_lowestX_lu[1]),
                            util::roundToInt(maxX_lowestY_lu[2]), util::roundToInt(maxY_lowestX_lu[2]));

       // Second Big File

        Array<T,3> maxX_lowestY_2(1765, 175, 8); // extreme right
        Array<T,3> maxY_lowestX_2(1561, 610, 32); // extreme left

         // Array bigWholeParty for velocity
        Array<T,3> maxX_lowestY_lu_2((maxX_lowestY_2-location)/dx); // to lattice Units
        Array<T,3> maxY_lowestX_lu_2((maxY_lowestX_2-location)/dx);  // to lattice Units


        Box3D yz_vtkWhole_2 (util::roundToInt(maxY_lowestX_lu_2[0]), util::roundToInt(maxX_lowestY_lu_2[0]),
                             util::roundToInt(maxX_lowestY_lu_2[1]), util::roundToInt(maxY_lowestX_lu_2[1]),
                             util::roundToInt(maxX_lowestY_lu_2[2]), util::roundToInt(maxY_lowestX_lu_2[2]));

        if(vtk1)
        {
            writeVtkPlus(lattice,
                     particles,
                     volumeFraction,
                     flagMatrix,
                     flagMatrixPlot,
                     particlesDynamicsPlot,
                     erosionDepositionPlot,
                     distanceFunction,
                     wss,
                     temporalMeanWss,
                     //yz_vtkWhole,
                     lattice.getBoundingBox(),
                     "bigVolume",
                     location,
                     dx,
                     dt,
                     iter);
        }

        if(vtk2){
            writeVtkPlus2(lattice, particles, volumeFraction,flagMatrix, flagMatrixPlot, particlesDynamicsPlot,
                 erosionDepositionPlot, distanceFunction, wss, temporalMeanWss, yz_vtkWhole_2, "bigVolume_2" , location, dx, dt, iter);
        }
    }

    /*
    writeVtkPlus(lattice, particles, volumeFraction,flagMatrix, flagMatrixPlot, particlesDynamicsPlot,
		 erosionDepositionPlot, distanceFunction, lattice.getBoundingBox(), "bigVolume" , location, dx, dt, iter);
*/
}

std::string meshFileNameVisu;
// Read the user input XML file provided at the command-line.
void readParameters(XMLreader const& document)
{
    std::string meshFileNameOne;
    std::string meshFileNameTwo;
    document["geometry"]["meshOne"].read(meshFileNameOne); // after erosion
    document["geometry"]["meshTwo"].read(meshFileNameTwo); // before erosion
    document["geometry"]["meshVisu"].read(meshFileNameVisu); // before erosion
    document["geometry"]["inletDiameter"].read(userDefinedInletDiameter);
    document["geometry"]["Xi0"].read(Xi0);
    document["geometry"]["Xi1"].read(Xi1);
    document["geometry"]["Yi0"].read(Yi0);
    document["geometry"]["Yi1"].read(Yi1);
    document["geometry"]["Zi0"].read(Zi0);
    document["geometry"]["Zi1"].read(Zi1);
    document["geometry"]["Xo0"].read(Xo0);
    document["geometry"]["Xo1"].read(Xo1);
    document["geometry"]["Yo0"].read(Yo0);
    document["geometry"]["Yo1"].read(Yo1);
    document["geometry"]["Zo0"].read(Zo0);
    document["geometry"]["Zo1"].read(Zo1);
    document["geometry"]["waterLevelFactor"].read(waterLevelFactor);

    document["particles"]["particleNumber"].read(particleNumber);
    document["particles"]["stressThreshold"].read(stressThreshold);
    document["particles"]["criticalTime"].read(criticalTime);
    document["particles"]["criticalNumber"].read(criticalNumber);
    document["particles"]["kappa"].read(kappa);
    document["particles"]["beginParticleDynamics"].read(beginParticleDynamics);
    document["particles"]["callDepositionErosion"].read(callDepositionErosion);
    document["particles"]["settlingVelocityLatticeUnits"].read(settlingVelocityLatticeUnits);

    document["numerics"]["dt"].read(dt);
    document["numerics"]["gPhys"].read(gPhys);
    document["numerics"]["referenceDirection"].read(referenceDirection);
    document["numerics"]["referenceResolution"].read(referenceResolution);
    document["numerics"]["maxVelocityInletPhys"].read(maxVelocityInletPhys);
    document["numerics"]["nuRU"].read(nuRU);
    document["numerics"]["fluidDensity"].read(fluidDensity);
    // document["numerics"]["inletFluxPhys"].read(inletFluxPhys);

    document["simulation"]["simTime"].read(simTime);
    document["simulation"]["cSmago"].read(cSmago);
    document["simulation"]["useParticles"].read(useParticles);
    document["simulation"]["dataWrite"].read(dataWrite);
    document["simulation"]["vtk1"].read(vtk1);
    document["simulation"]["vtk2"].read(vtk2);
    document["simulation"]["distanceSTL"].read(distanceSTL);
    document["simulation"]["interfaceSTL"].read(interfaceSTL);
    document["simulation"]["onCluster"].read(onCluster);
    document["simulation"]["minStatIter"].read(minStatIter);

    // At this part, the surface geometry of the user (as given by the user in
    //   the form of an ASCII or binary STL file) is read into a data structure
    //   comprised by a set of triangles. The DBL constant means that double
    //   precision accuracy will be used (generally the recommended choice).
    outerTriangleSet = new TriangleSet<T>(meshFileNameOne, DBL); // outer stl --> after complete erosion geometry
    innerTriangleSet = new TriangleSet<T>(meshFileNameTwo, DBL); // inner stl --> initial bed river condition

    pcout << "Use particles: " << useParticles << std::endl;
    pcout << "Write interface: " << interfaceSTL << std::endl;
}


void writeStlOfDomain(Box3D domain, T definition, T dx, Array<T,3> location, std::string filename)
{
    plint nx = domain.getNx();
    if(nx < 3)
    {
        pcout << filename << " : redimension, nx was equal to " << nx << std::endl;
        nx = 3;
    }
    plint ny = domain.getNy();
    if(ny < 3)
    {
        pcout << filename << " : redimension, ny was equal to " << ny << std::endl;
        ny = 3;
    }
    plint nz = domain.getNz();
    if(nz < 3)
    {
        pcout << filename << " : redimension, nz was equal to " << nz << std::endl;
        nz = 3;
    }
    MultiScalarField3D<T> field(nx, ny, nz);
    setToConstant(field, field.getBoundingBox().enlarge(-1), 1.);
    pcout << "Writing stl " + filename << std::endl;
    pcout << field.getBoundingBox().x0 << " " << field.getBoundingBox().x1 << std::endl;
    pcout << field.getBoundingBox().y0 << " " << field.getBoundingBox().y1 << std::endl;
    pcout << field.getBoundingBox().z0 << " " << field.getBoundingBox().z1 << std::endl;

    // we construct triangles that span the bounding box
    std::vector<T> isoLevels;
    isoLevels.push_back((T) definition);
    typedef TriangleSet<T>::Triangle Triangle;
    std::vector<Triangle> triangles;
    isoSurfaceMarchingCube(triangles, field, isoLevels, field.getBoundingBox());

    // first option: we make a TriangleSet
    TriangleSet<T> triangleSet(triangles);
    //triangleSet.scale(dx);
    triangleSet.translate(Array<T,3>(domain.x0, domain.y0, domain.z0));
    triangleSet.writeBinarySTL(filename);

    // second option : DEFscaledMesh
    /*
    Dot3D loc = Dot3D(location[0], location[1], location[2]);
    plint dz = waterLevel;
    plint refDirection = 2;
    DEFscaledMesh<T>* defMesh = new DEFscaledMesh<T>(TriangleSet<T>(triangles),
                                                     dz,
                                                     refDirection,
                                                     margin,
                                                     loc);
    TriangleBoundary3D<T> boundary(*defMesh);
    boundary.getMesh().writeAsciiSTL(filename);
    */
}

// inlet and outlet for fluid lattice.
void defineInletOutletForFluid (MultiBlockLattice3D<T, DESCRIPTOR>& lattice,
                                Array<T,3> location,
                                T dx)
{
    pcout << "Location: " << location[0] << " " << location[1] << " " << location[2] << std::endl;
    Box3D bb = lattice.getBoundingBox();
    pcout << "      Lattice BBx : " << bb.x0 << " " << bb.x1 << std::endl;
    pcout << "      Lattice BBy : " << bb.y0 << " " << bb.y1 << std::endl;
    pcout << "      Lattice BBz : " << bb.z0 << " " << bb.z1 << std::endl;

    pcout << "Inlet surface: " << inletSurfPhys << "m2" << std::endl;
    pcout << "Outlet volume: " << outletVolPhys << "m3" << std::endl;

    //inletDomain.x0 = lattice.getBoundingBox().x1 - margin;
    //inletDomain.x1 = lattice.getBoundingBox().x1 - margin;
    // inletDomain.x0 = util::roundToInt(Xi/dx - location[0]/dx);
    // inletDomain.x1 = util::roundToInt(Xi/dx - location[0]/dx);
    // inletDomain.y0 = util::roundToInt(Yi0/dx - location[1]/dx);
    // inletDomain.y1 = util::roundToInt(Yi1/dx - location[1]/dx);
    // inletDomain.z0 = util::roundToInt(Zi0/dx - location[2]/dx);
    // inletDomain.z1 = util::roundToInt(Zi1/dx - location[2]/dx);

    // inletDomain.x0 = util::roundToInt(Xi);
    // inletDomain.x1 = util::roundToInt(Xi);
    // inletDomain.y0 = util::roundToInt(Yi0);
    // inletDomain.y1 = util::roundToInt(Yi1);
    // inletDomain.z0 = util::roundToInt(Zi0);
    // inletDomain.z1 = util::roundToInt(Zi1);

    const plint nx = lattice.getNx();
    const plint ny = lattice.getNy();
    const plint nz = lattice.getNz();

    inletDomain.x0 = (plint) (Xi0 * nx);
    inletDomain.x1 = (plint) (Xi1 * nx);

    inletDomain.y0 = (plint) (Yi0 * ny);
    inletDomain.y1 = (plint) (Yi1 * ny);

    inletDomain.z0 = (plint) (Zi0 * nz);
    inletDomain.z1 = (plint) (Zi1 * nz);

    pcout << "nx ny nz " << nx << " " << ny << " " << nz << std::endl;
    pcout << "xi yi zi = "  << Xi0 << " " << Xi1 << std::endl;
    pcout << "inletDomain.x0 = "  << inletDomain.x0 << "; inletDomain.x1 = " << inletDomain.x1 << std::endl;
    pcout << "inletDomain.y0 = "  << inletDomain.y0 << "; inletDomain.y1 = " << inletDomain.y1 << std::endl;
    pcout << "inletDomain.z0 = "  << inletDomain.z0 << "; inletDomain.z1 = " << inletDomain.z1 << std::endl;

    if (inletDomain.x0 < bb.x0)
    {
        pcout << "Correction of inlet x0 " << inletDomain.x0 << std::endl;
        inletDomain.x0 = bb.x0;
    }

    if (inletDomain.x1 > bb.x1)
    {
        pcout << "Correction of inlet x1 " << inletDomain.x1 << std::endl;
        inletDomain.x1 = bb.x1;
    }

    if (inletDomain.y0 < bb.y0)
    {
        pcout << "Correction of inlet y0 " << inletDomain.y0 << std::endl;
        inletDomain.y0 = bb.y0;
    }

    if (inletDomain.y1 > bb.y1)
    {
        pcout << "Correction of inlet y1 " << inletDomain.y1 << std::endl;
        inletDomain.y1 = bb.y1;
    }

    if (inletDomain.z0 < bb.z0)
    {
        pcout << "Correction of inlet z0 " << inletDomain.z0 << std::endl;
        inletDomain.z0 = bb.z0;
    }

    if (inletDomain.z1 > bb.z1)
    {
        pcout << "Correction of inlet z1 " << inletDomain.z1 << std::endl;
        inletDomain.z1 = bb.z1;
    }

    PLB_ASSERT(inletDomain.x0 <= inletDomain.x1);
    PLB_ASSERT(inletDomain.y0 <= inletDomain.y1);
    PLB_ASSERT(inletDomain.z0 <= inletDomain.z1);
    PLB_ASSERT(inletDomain.x0==inletDomain.x1 || // assert inlet domain is a surface
                inletDomain.y0==inletDomain.y1 ||
                inletDomain.z0==inletDomain.z1);


    pcout << "inletDomain.x0 and inletDomain.x1 = " << inletDomain.x0 << " " << inletDomain.x1 << std::endl;
    pcout << "inletDomain.y0 = "  << inletDomain.y0 << "; inletDomain.y1 = " << inletDomain.y1 << std::endl;
    pcout << "inletDomain.z0 = "  << inletDomain.z0 << "; inletDomain.z1 = " << inletDomain.z1 << std::endl;

    // this region is the one where we apply the removeMass algorithm
    outletDomain.x0 = (plint) (Xo0 * nx);
    outletDomain.x1 = (plint) (Xo1 * nx);

    outletDomain.y0 = (plint) (Yo0 * ny);
    outletDomain.y1 = (plint) (Yo1 * ny);

    outletDomain.z0 = (plint) (Zo0 * nz);
    outletDomain.z1 = (plint) (Zo1 * nz);

    pcout << " *** Before Correction *** " << std::endl;
    pcout << "outletDomain.x0 = "  << outletDomain.x0 << "; outletDomain.x1 = " << outletDomain.x1 << std::endl;
    pcout << "outletDomain.y0 = "  << outletDomain.y0 << "; outletDomain.y1 = " << outletDomain.y1 << std::endl;
    pcout << "outletDomain.z0 = "  << outletDomain.z0 << "; outletDomain.z1 = " << outletDomain.z1 << std::endl;

    if (outletDomain.x0 < bb.x0)
    {
        pcout << "Correction of outlet x0 " << outletDomain.x0 << std::endl;
        outletDomain.x0 = bb.x0;
    }

    if (outletDomain.x1 > bb.x1)
    {
        pcout << "Correction of inlet x1 " << outletDomain.x1 << std::endl;
        outletDomain.x1 = bb.x1;
    }

    if (outletDomain.y0 < bb.y0 )
    {
        pcout << "Correction of outlet y0 " << outletDomain.y0 << std::endl;
        outletDomain.y0 = bb.y0;
    }

    if (outletDomain.y1 > bb.y1 )
    {
        pcout << "Correction of outlet y1 " << outletDomain.y1 << std::endl;
        outletDomain.y1 = bb.y1;
    }

    if (outletDomain.z0 < bb.z0 )
    {
        pcout << "Correction of outlet z0 " << outletDomain.z0 << std::endl;
        outletDomain.z0 = bb.z0;
    }

    if (outletDomain.z1 > bb.z1 )
    {
        pcout << "Correction of outlet z1 " << outletDomain.z1 << std::endl;
        outletDomain.z1 = bb.z1;
    }

    PLB_ASSERT(inletDomain.x0 <= inletDomain.x1);
    PLB_ASSERT(inletDomain.y0 <= inletDomain.y1);
    PLB_ASSERT(inletDomain.z0 <= inletDomain.z1);

    pcout << " *** After Correction *** " << std::endl;
    pcout << "outletDomain.x0 = "  << outletDomain.x0 << "; outletDomain.x1 = " << outletDomain.x1 << std::endl;
    pcout << "outletDomain.y0 = "  << outletDomain.y0 << "; outletDomain.y1 = " << outletDomain.y1 << std::endl;
    pcout << "outletDomain.z0 = "  << outletDomain.z0 << "; outletDomain.z1 = " << outletDomain.z1 << std::endl;

    PLB_ASSERT(inletDomain.x0 <= inletDomain.x1);
    PLB_ASSERT(inletDomain.y0 <= inletDomain.y1);
    PLB_ASSERT(inletDomain.z0 <= inletDomain.z1);

    pcout << "lattice.getBoundingBox().x0  = "  << lattice.getBoundingBox().x0 << "; lattice.getBoundingBox().x1 = " << lattice.getBoundingBox().x1 << std::endl;

    plint riverLenght = lattice.getBoundingBox().x1 - lattice.getBoundingBox().x0;
    pcout << "riverLenght = " << riverLenght << std::endl;

    // ??? utilise outletDomain. Devrait etre fait apres le defineoutlet????
    depositionErosionZone = lattice.getBoundingBox();
    depositionErosionZone.x0 = outletDomain.x1 + 1;
    depositionErosionZone.x1 = inletDomain.x1 -20 ;

    pcout << "depositionErosionZone.x0 = "  << depositionErosionZone.x0 << "; depositionErosionZone.x1 = " << depositionErosionZone.x1 << std::endl;
    pcout << "depositionErosionZone.y0 = "  << depositionErosionZone.y0 << "; depositionErosionZone.y1 = " << depositionErosionZone.y1 << std::endl;
    pcout << "depositionErosionZone.z0 = "  << depositionErosionZone.z0 << "; depositionErosionZone.z1 = " << depositionErosionZone.z1 << std::endl;

/*
    MultiScalarField3D<T> odfield(outletDomain.getNx(), outletDomain.getNy(), outletDomain.getNz());
    setToConstant(odfield, odfield.getBoundingBox(), 1.);
*/
    writeStlOfDomain(outletDomain, 0.01, dx, location, "a_outletDomain.stl");
    //writeStlOfDomain(lattice.getBoundingBox(), 0.1, dx, location, myPath + "a_latticeDomain.stl");
    writeStlOfDomain(inletDomain, 0.01, dx, location, "a_inletDomain.stl");

}



void defineInletOutletPerParticles (MultiBlockLattice3D<T, DESCRIPTOR2>& particles, Array<T,3> location, T dx)
{
   inletDomainParticles = inletDomain.shift(-5, 0, 0);
   outletDomainParticles = outletDomain.shift(0, 0, 0);
   outletDomainParticles.y0 = outletDomain.y0 + 2;
   outletDomainParticles.y1 = outletDomain.y0 + 2;
   writeStlOfDomain(outletDomainParticles, 0.01, dx, location, "outletDomainParticles.stl");
   writeStlOfDomain(inletDomain, 0.01, dx, location, "inletDomainParticles.stl");
}


void particlesBoundaryConditionAndInitialization (
        MultiBlockLattice3D<T, DESCRIPTOR2>& particles,
        OnLatticeAdvectionDiffusionBoundaryCondition3D<T,DESCRIPTOR2>& adBoundaryCondition,
				 Box3D inletDomainParticles, Box3D outletDomainParticles, T particleNumber, T iniParts)
{

    initializeAtEquilibrium(particles, particles.getBoundingBox(), particleNumber, Array<T,3>(0.,0.,0.) );
    initializeAtEquilibrium(particles, inletDomainParticles, particleNumber/5, Array<T,3>(0.,0.,0.) );
    initializeAtEquilibrium(particles, outletDomainParticles, T(), Array<T,3>(0.,0.,0.) );


    adBoundaryCondition.addTemperatureBoundary0P(inletDomainParticles,  particles);
    adBoundaryCondition.addTemperatureBoundary0N(outletDomainParticles, particles, boundary::neumann);

    pcout << "particleNumber = " << particleNumber << std::endl;

    setBoundaryDensity(particles, inletDomainParticles, particleNumber  );
    setBoundaryDensity(particles, outletDomainParticles, 0.  );

    setExternalVector(particles, particles.getBoundingBox(),
                      DESCRIPTOR2<T>::ExternalField::velocityBeginsAt, Array<T,3>(0.,0.,0.));

    particles.initialize();

}




int main(int argc, char* argv[])
{
    plbInit(&argc, &argv);

    string paramXmlFileName;
    try {
        global::argv(1).read(paramXmlFileName);
    }
    catch (PlbIOException& exception) {
        pcout << "Wrong parameters; the syntax is: "
              << (std::string)global::argv(0) << " parameter-input-file.xml" << std::endl;
        return -1;
    }

    // Read the parameter XML input file. (Lots of comments are included there too).
    XMLreader document(paramXmlFileName);
    pcout << "xml param opened." << std::endl;
    readParameters(paramXmlFileName);
    global::timer("global").start();

    pcout << "Interface: " << interfaceSTL << endl;

    if (onCluster){
        global::IOpolicy().activateParallelIO(true);
    }
    else{
        global::IOpolicy().activateParallelIO(false);
    }

    // Requirement: margin>=borderWidth.
    pcout << "XXX" << std::endl;
    // new stuff that allow us to use two different stl geomtries
    //outerTriangleSet comes from meshFileName1, so AFTER erosion.
    DEFscaledMesh<T>* outerDefMesh = new DEFscaledMesh<T>(*outerTriangleSet,
                                                            referenceResolution,
                                                            referenceDirection,
                                                            margin,
                                                            extraLayer);
    //DEFscaledMesh<T>* outerDefMesh = new DEFscaledMesh<T>(*outerTriangleSet);
    pcout << "A" << std::endl;

    TriangleBoundary3D<T> outerBoundary(*outerDefMesh);

    pcout << "B" << std::endl;
    delete outerDefMesh;

    // outerBoundary.getMesh().inflate();  // should I or I shouldn't inflate this stuff?
    // outerBoundary.getMesh().writeAsciiSTL(myPath + "scaled_outerRhoneGeometry.stl");

    pcout <<"B.1" << std::endl;
    T dx 	= outerBoundary.getDx();
    maxVelocityInletLB = maxVelocityInletPhys * dt / dx;
    T reynoldsNumber = maxVelocityInletPhys*userDefinedInletDiameter/nuRU;
    T nuLB_	= nuRU*dt/(dx*dx);
    //T tau = nuRU * DESCRIPTOR<T>::invCs2 * dt / (dx*dx) + 0.5;
    //T omegaYann = 1./tau;
    T omega 	= 1./(3.*nuLB_+0.5);
    Array<T,3> location(outerBoundary.getPhysicalLocation()); // Compute location from the mesh
    T gLB = -1. * gPhys * dt * dt / dx;
    externalForce = Array<T,3>(0., 0., gLB);

    //T Zo0 = Zi0;
    //T Zo1 = Zi0 + waterLevel + 5;
    // inletSurfPhys = (inletDomain.x1 - inletDomain.x0 + 1) *
    //                 (inletDomain.y1 - inletDomain.y0 + 1) *
    //                 (inletDomain.z1 - inletDomain.z0 + 1);

    // outletVolPhys = (outletDomain.z1 - outletDomain.z0 + 1) *
    //                 (outletDomain.y1 - outletDomain.y0 + 1) *
    //                 (outletDomain.z1 - outletDomain.z0 + 1);
    // pcout << "outletVolLB: " << outletVolPhys << " m3" << std::endl;
    // outletVolPhys *= dx*dx*dx;
    //
    // // pcout << "inletSurfPhys: " << inletSurfPhys << " m2" << std::endl;
    // pcout << "outletVolPhys: " << outletVolPhys << " m3" << std::endl;
    // pcout << "      box used (inlet): " << Xi0 << " " << Xi1 << " " << Yi0 << " " << Yi1 << " " << Zi0 << " " << Zi1 << std::endl;
    // pcout << "      box used (outlet): " << Xo0 << " " << Xo1 << " " << Yo0 << " " << Yo1 << " " << Zo0 << " " << Zo1 << std::endl;
    //
    // // T maxVelVerif = inletFluxPhys / inletSurfPhys;
    // // pcout << "Inlet max vel verif: " << maxVelVerif << "m/s" << std::endl;
    //
    // removeMassIter = outletVolPhys / outletFluxPhys;
    // removeMassIter /= dt;
    // pcout << "Remove mass Iter: " << removeMassIter << std::endl;
    // PLB_ASSERT(removeMassIter > 0);
    // /*
    // Ideally, we would like to remove as less mass as possible as much often
    // as possible (each iteration), in order to keep the discontinuites due to
    // mass removal low.
    // */

    removeMassIter = 1;


    Array<T,3> settlingVelocityVector (0.,0., -settlingVelocityLatticeUnits);

    pcout << "Computed Re=" << reynoldsNumber << std::endl;
    pcout << "uLB=" << maxVelocityInletLB << std::endl;
    pcout << "nuLB=" << nuLB_ << std::endl;
    pcout << "nuRU=" << nuRU << std::endl;
    pcout << "tau=" << 1./omega << std::endl;
    pcout << "omega=" << omega << std::endl;
	pcout << "dx=" << dx << std::endl;
	pcout << "dt=" << dt << std::endl;

    // for particles diffusion
    T kappa_lu = kappa*dt/(dx*dx);
    T omegaAD = (T)1/(kappa_lu * DESCRIPTOR2<T>::invCs2 + 0.5);

	pcout << "kappa =" << kappa << std::endl;
	pcout << "kappa_lu =" << kappa_lu << std::endl;
	pcout << "omegaAD =" << omegaAD << std::endl;
	pcout << "tauAD =" << 1./omegaAD << std::endl;

    // The rhone simulation is an interior (as opposed to exterior) flow problem. For
    //   this reason, the lattice nodes that lay inside the computational domain must
    //   be identified and distinguished from the ones that lay outside of it. This is
    //   handled by the following voxelization process.
    //   pk doivent-ils etres distingues???? ==> performance



    // BUILDING BOUNCEBACKFLAGMATRIX *******************************************************************************
    // in what follows, outerFlagMatrix and innerFlagMatrix are juste a way to build bounceBackFlagMatrix properly.
    // then bounceBackFlagMatrix can be used to initialize the simulation.
    const int flowType = voxelFlag::inside;
    // create the domain corresponding to stl"after erosion"
    VoxelizedDomain3D<T> erodableVoxelizedDomain (outerBoundary,
                                                  flowType,
                                                  extraLayer,
                                                  borderWidth,
                                                  envelopeWidth,
                                                  blockSize);

	pcout << getMultiBlockInfo(erodableVoxelizedDomain.getVoxelMatrix()) << std::endl;

    //flag matrix for the "after erosion" domain : 2 for not-erodable wall, 1 for erodable wall
    MultiScalarField3D<int> outerFlagMatrix((MultiBlock3D&) erodableVoxelizedDomain.getVoxelMatrix());
    setToConstant(outerFlagMatrix, outerFlagMatrix.getBoundingBox(), 2); //not-erodable wall everywhere...
    setToConstant(outerFlagMatrix, erodableVoxelizedDomain.getVoxelMatrix(),
                  voxelFlag::inside, outerFlagMatrix.getBoundingBox(), 1); // and then, erodable wall at "inside" nodes.


    // and now I create the voxelized domain for the inner stl

    // Come in Jonas code per micro stent
    //innerTriangleSet comes from meshFileName2
    DEFscaledMesh<T>* innerDefMesh = new DEFscaledMesh<T>(*innerTriangleSet,  0, 0, margin, extraLayer);
    // put the inner at the same location as the outer
    innerDefMesh->getMesh().translate(-location);
    // scale the inner at the same scale as the outer (remember that dx is defined in function of outer)
    innerDefMesh->getMesh().scale(1./dx);
    TriangleBoundary3D<T> innerBoundary(*innerDefMesh);
    innerBoundary.getMesh().writeAsciiSTL(outputDir + "inner_bedRiverGeometry.stl");
    outerBoundary.getMesh().writeAsciiSTL(outputDir + "outer_bedRiverGeometry.stl");


    //write domain without roof for visualization purpose only//////////////////
    visuTriangleSet = new TriangleSet<T>(meshFileNameVisu, DBL);
    DEFscaledMesh<T>* visuDefMesh = new DEFscaledMesh<T>(*visuTriangleSet, 0, 0, margin, extraLayer);
    visuDefMesh->getMesh().translate(-location);
    visuDefMesh->getMesh().scale(1./dx);
    TriangleBoundary3D<T> visuBoundary(*visuDefMesh);
    visuBoundary.getMesh().writeAsciiSTL(outputDir + "visu_bedRiverGeometry.stl");
    ////////////////////////////////////////////////////////////////////////////


    pcout << std::endl << "Voxelizing the inner bed river domain" << std::endl;
    Array<T,2> xRange, yRange, zRange;
    plint layer = margin + extraLayer;
    innerDefMesh->getMesh().computeBoundingBox(xRange, yRange, zRange);
    delete innerDefMesh;
    Box3D innerBoundingBox(util::roundToInt(xRange[0]) - layer, util::roundToInt(xRange[1] + layer),
                           util::roundToInt(yRange[0]) - layer, util::roundToInt(yRange[1] + layer),
                           util::roundToInt(zRange[0]) - layer, util::roundToInt(zRange[1] + layer));
    const int innerFlowType = voxelFlag::inside;

    VoxelizedDomain3D<T> innerVoxelizedDomain (innerBoundary, innerFlowType, innerBoundingBox,
                                               borderWidth, extendedEnvelopeWidth, blockSize);
    pcout << getMultiBlockInfo(innerVoxelizedDomain.getVoxelMatrix()) << std::endl;

    //flag matrix for the "before erosion" domain : 1 for erodable wall, 0 for bulk (=noWall, either fluid or empty)
    MultiScalarField3D<int> innerFlagMatrix_tmp((MultiBlock3D&) innerVoxelizedDomain.getVoxelMatrix());

    setToConstant(innerFlagMatrix_tmp, innerFlagMatrix_tmp.getBoundingBox(), 1); // erodableWall everywhere...

    setToConstant(innerFlagMatrix_tmp, innerVoxelizedDomain.getVoxelMatrix(), // ...and then, bulk at "inside" nodes
                  voxelFlag::inside, innerFlagMatrix_tmp.getBoundingBox(), 0);

    //putDam(0.0746, 0.00421, 0.791, 0.284, innerFlagMatrix_tmp, innerVoxelizedDomain, 1); // 1 --> 2

    // et voila ... the magic moment. Where I create a innerFlagMatrix that has the same size of
    // the outer flag matrix... but contains the infos about the inner stl geometry
    MultiScalarField3D<int> innerFlagMatrix((MultiBlock3D&) outerFlagMatrix);
    setToConstant(innerFlagMatrix, innerFlagMatrix.getBoundingBox(), 1);// wall everywhere...
    copyNonLocal(innerFlagMatrix_tmp, innerFlagMatrix, innerFlagMatrix.getBoundingBox()); //...and then fill the rest with tmp matrix values

    pcout << "Number of initial inside the bed river geometry: " << computeSum(innerFlagMatrix) << std::endl;

    // This is the mask matrix. Where a value of 1 exists, a bounce back node will be placed in the lattice.
    MultiScalarField3D<int> bounceBackFlagMatrix((MultiBlock3D&) outerFlagMatrix);
    setToConstant(bounceBackFlagMatrix, bounceBackFlagMatrix.getBoundingBox(), 2); // not erodable wall everywhere

    pcout << "A BLOP" << std::endl;

    // the result of the multiplication is that bulk wins all, then notErodableWall, then erodableWall.
    // the result of the multiplication outer * inner is stored into bounce.
    multiply(outerFlagMatrix, innerFlagMatrix, bounceBackFlagMatrix, bounceBackFlagMatrix.getBoundingBox());
    pcout << "B BLOP" << std::endl;
    // Normally we have:
    //                1. a region where the bounceBackFlagMatrix = 2 ---> NoDynamics Region, notErodableWall;
    //			      2. a region where the bounceBackFlagMatrix = 1 ---> BounceBack Region, erodableWall;
    //			      3. a region where the bounceBackFlagMatrix = 0 ---> FreeSurfaceFluid Region with initial empty condition;
    // devrait y avoir des regions avec 4 a cause de multiplication du outer avec bounceBack!
    Dynamics<T,DESCRIPTOR>* dynamics = 0;
    dynamics = new SmagorinskyBGKdynamics<T,DESCRIPTOR>(omega, cSmago);
    pcout << "C BLOP" << std::endl;

     /////////////////////////////////  Initialization for lattice and freeSurface utilities BEGIN /////////////////////////////

    plint envelopeWidth = 1;
    MultiBlockLattice3D<T,DESCRIPTOR>* fluid = generateMultiBlockLattice<T,DESCRIPTOR>(erodableVoxelizedDomain.getVoxelMatrix(), envelopeWidth, dynamics->clone()).release();
    Group3D fsFields(fluid, "fluid1");

    pcout << "Domain size : " << outerFlagMatrix.getNx() << " " << outerFlagMatrix.getNy() <<  " " << outerFlagMatrix.getNz() << std::endl;
    /*
    T contactAngle = 90.;
    T surfaceTension = 0.07;
    */
    const T contactAngle = -1; //deactivate surface tension algorithm
    const T surfaceTension = 0.;
    pcout << "Before setup" << std::endl;
    FreeSurfaceSetup<T,DESCRIPTOR> fsSetup(fsFields, *dynamics, rhoEmpty, surfaceTension, contactAngle, externalForce);
    pcout << "After setup" << std::endl;

    pcout << "a" << std::endl;
    fsSetup.createFreeSurfaceFields();
    pcout << "b" << std::endl;
    fsSetup.createFreeSurfaceProcessors();

    pcout << "Lattice:" << std::endl;
    Box3D lbox = fsFields.getBoundingBox();
    pcout << "X:" << lbox.x0 << " " << lbox.x1 << std::endl;
    pcout << "Y:" << lbox.y0 << " " << lbox.y1 << std::endl;
    pcout << "Z:" << lbox.z0 << " " << lbox.z1 << std::endl;
    defineInletOutletForFluid(fsFields.getLattice<T,DESCRIPTOR>("fluid1"), location, dx);

    pcout << "E BLOP" << std::endl;

    std::vector<MultiBlock3D* > iniFlagStatusFreeSurface;
    iniFlagStatusFreeSurface.push_back(&fsFields.getLattice<T,DESCRIPTOR>("fluid1"));
    iniFlagStatusFreeSurface.push_back(&fsFields.getScalar<int>("flag"));
    iniFlagStatusFreeSurface.push_back(&bounceBackFlagMatrix);

    pcout << "F BLOP" << std::endl;

    setExternalScalar(fsFields.getLattice<T,DESCRIPTOR>("fluid1"),
                      fsFields.getBoundingBox(),
                      DESCRIPTOR<T>::ExternalField::nodeNatureBeginsAt,
                      (T)NodeNature::bulk);
    pcout << "G BLOP" << std::endl;

    setToConstant(fsFields.getScalar<int>("flag"), fsFields.getBoundingBox(), (int)freeSurfaceFlag::wall);
    pcout << "H BLOP" << std::endl;

    // set flag and nodeNature
    // IniFreeSurfaceFlagFunctional(waterLevel, inletPouring) : put water, set erodable and not erodable walls.
    /*
    applyProcessingFunctional(new IniFreeSurfaceFlagFunctional3D<T,DESCRIPTOR>(inletDomain.z1, inletDomain),
                              fsFields.getBoundingBox(),
                              iniFlagStatusFreeSurface);
    */
    // plint domainHeight = inletDomain.z1 - inletDomain.z0;
    plint iniWaterLevel = (plint) (waterLevelFactor*lbox.getNz());
    pcout << "IniWaterLevel = " << iniWaterLevel << " : " << lbox.z0 << " " << lbox.z1 << " " << waterLevelFactor << std::endl;
    //iniWaterLevel fills with water each empty (non bounce-back) cell of the domain below iniWaterLevel.
    applyProcessingFunctional(new IniFreeSurfaceFlagFunctional3D<T,DESCRIPTOR>(iniWaterLevel, inletDomain),
                              fsFields.getBoundingBox(),
                              iniFlagStatusFreeSurface);

    fsSetup.defaultInitialize();

    MultiScalarField3D<T> flagMatrixPlot((MultiBlock3D&)erodableVoxelizedDomain.getVoxelMatrix());

    MultiScalarField3D<T>* particlesDynamicsPlot=0;
    MultiScalarField3D<T>* erosionDepositionPlot=0;

    if (useParticles)
    {
        particlesDynamicsPlot = new MultiScalarField3D<T>((MultiBlock3D&)erodableVoxelizedDomain.getVoxelMatrix());
        erosionDepositionPlot = new MultiScalarField3D<T>((MultiBlock3D&)erodableVoxelizedDomain.getVoxelMatrix());
        setToConstant(*erosionDepositionPlot, erosionDepositionPlot->getBoundingBox(), T(1));
    }
    // topLid position (used for particles lattice. A bounce back dynamic is applied on it but we fix a number of particles = 0)
    // Elsewhere on the river bed, the number of particles is representative of a deposit lattice node.
    Box3D topLid = fsFields.getBoundingBox();
    topLid.z0 = topLid.z1-margin-2;

    pcout << "topLid.x0 = "  << topLid.x0 << "; topLid.x1 = " << topLid.x1 << std::endl;
    pcout << "topLid.y0 = "  << topLid.y0 << "; topLid.y1 = " << topLid.y1 << std::endl;
    pcout << "topLid.z0 = "  << topLid.z0 << "; topLid.z1 = " << topLid.z1 << std::endl;

    setExternalScalar(fsFields.getLattice<T,DESCRIPTOR>("fluid1"), fsFields.getBoundingBox(), DESCRIPTOR<T>::ExternalField::timeBeginsAt, (T)0);

    setExternalScalar(fsFields.getLattice<T,DESCRIPTOR>("fluid1"), fsFields.getBoundingBox(), DESCRIPTOR<T>::ExternalField::omegaBeginsAt, omega);

    // set nextToWall node nature for cells that are next to a bounce-back cell and that arent bounce-back themselves.
    applyProcessingFunctional(new DefineBulkStatusFunctional3D<T,DESCRIPTOR>(inletDomain.z1+2), fsFields.getBoundingBox(),
                              fsFields.getLattice<T,DESCRIPTOR>("fluid1"), fsFields.getScalar<int>("flag"));

    //  Thinks in terms of volume flux ...
    // T totalMomentumAtSlit = T(maxVelocityInletLB);
    // Array<T,3> momentumAtInlet ((T)0, -totalMomentumAtSlit, (T)0);
    // Array<T,3> momentumAtInlet (-totalMomentumAtSlit, (T)0, (T)0);
    Array<T,3> velocityAtInlet ((T)0, -T(maxVelocityInletLB), (T)0);
    integrateProcessingFunctional(new PouringLiquid3D<T,DESCRIPTOR>(dynamics, velocityAtInlet), // CF multiPhysics/freesurfaceboundarycondition for Pouring
                                  inletDomain,
                                  fsSetup.freeSurfaceArgs);
    pcout << "Velocity inlet LB = " << maxVelocityInletLB << std::endl;

    /////////////////////////////////  Initialization for lattice and freeSurface utilities END /////////////////////////////



    Dynamics<T,DESCRIPTOR2>* dynamicsParticles = 0;
    if (useParticles)
    {
        dynamicsParticles = new AdvectionDiffusionRLBdynamics<T,DESCRIPTOR2>(omegaAD);
    }

    std::auto_ptr<MultiBlockLattice3D<T,DESCRIPTOR2> > particles;
    if (useParticles)
    {
        particles = generateMultiBlockLattice<T,DESCRIPTOR2> (erodableVoxelizedDomain.getVoxelMatrix(),
                                                              envelopeWidth,
                                                              new NoDynamics<T,DESCRIPTOR2>() );

        defineInletOutletPerParticles(*particles, location, dx);
        particles->toggleInternalStatistics(true);
        particles->periodicity().toggleAll(false);
    }


    Box3D eliminateParticles;
    std::vector<MultiBlock3D* > multiBlockChangeADdynamics;

    if (useParticles)
    {

        // Initialization of AD dyamics for bulk only
        applyProcessingFunctional(new IniDynamicsPerRiverFunctional3D<T,DESCRIPTOR2>(dynamicsParticles),
                                  particles->getBoundingBox(),
                                  *particles, fsFields.getScalar<int>("flag"));

        // Initialization of AD dyamics inner border (for all erodable or not erodable walls that have neighbour nextToWall or bulk, set dynamicsParticle, else NoDynamics).
        applyProcessingFunctional(new MakeThrombosisToParticlesFunctional3D<T,DESCRIPTOR,T,DESCRIPTOR2>(omegaAD,criticalNumber,dynamicsParticles),
                                        particles->getBoundingBox(), fsFields.getLattice<T,DESCRIPTOR>("fluid1"), *particles);

        OnLatticeAdvectionDiffusionBoundaryCondition3D<T,DESCRIPTOR2>*
            adBoundaryCondition = createLocalAdvectionDiffusionBoundaryCondition3D<T,DESCRIPTOR2>();

        particlesBoundaryConditionAndInitialization(*particles,
                                                    *adBoundaryCondition,
                                                    inletDomainParticles,
                                                    outletDomainParticles,
                                                    particleNumber,
                                                    iniParts);

        multiBlockChangeADdynamics.push_back(&fsFields.getLattice<T,DESCRIPTOR>("fluid1"));
        multiBlockChangeADdynamics.push_back(&*particles);
        multiBlockChangeADdynamics.push_back(&fsFields.getScalar<int>("flag"));

        eliminateParticles 	= fsFields.getBoundingBox();
        eliminateParticles.x0       = margin;
        eliminateParticles.x1 	    = fsFields.getBoundingBox().x1-margin-3; // pk -3?????????
        eliminateParticles.z0       = margin;
        eliminateParticles.z1       = fsFields.getBoundingBox().x1-margin;


        pcout << "eliminateParticles.x0 = "  << eliminateParticles.x0 << "; eliminateParticles.x1 = " << eliminateParticles.x1 << std::endl;
        pcout << "eliminateParticles.y0 = "  << eliminateParticles.y0 << "; eliminateParticles.y1 = " << eliminateParticles.y1 << std::endl;
        pcout << "eliminateParticles.z0 = "  << eliminateParticles.z0 << "; eliminateParticles.z1 = " << eliminateParticles.z1 << std::endl;
    }

    MultiScalarField3D<T>* wss;
    MultiScalarField3D<T>* temporalMeanWss;
    MultiScalarField3D<T>* temporalMeanWss_plot;
    MultiScalarField3D<T>* distanceFunction;
    std::vector<MultiBlock3D* > multiBlockForTemporalMeanWss;
    std::vector<MultiBlock3D* > particlesFluidAndWss;
    std::vector<MultiBlock3D* > multiBlockForUnknownAD;

    if (useParticles)
    {
        // get MultiScalarField containing the distance to wall
        distanceFunction = computeDistanceFunction(fsFields.getLattice<T,DESCRIPTOR>("fluid1"), fsFields.getLattice<T,DESCRIPTOR>("fluid1").getNy(), (T)1.0e-4).release();

        wss = new MultiScalarField3D<T>(fsFields.getLattice<T,DESCRIPTOR>("fluid1"));
        // compute and store wall shear stress into wss. Wss is computed using distanceFunction to define the normal.
        computeWallShearStress ( fsFields.getLattice<T,DESCRIPTOR>("fluid1"), *computeWallUnitNormal(*distanceFunction, fsFields.getBoundingBox()), *wss, wss->getBoundingBox() );

        temporalMeanWss = new MultiScalarField3D<T>(fsFields.getLattice<T,DESCRIPTOR>("fluid1"));
        setToConstant(*temporalMeanWss, temporalMeanWss->getBoundingBox(), T(0));

        temporalMeanWss_plot = new MultiScalarField3D<T>(fsFields.getLattice<T,DESCRIPTOR>("fluid1"));
        setToConstant(*temporalMeanWss_plot, temporalMeanWss_plot->getBoundingBox(), T(0));

        particlesFluidAndWss.push_back(&fsFields.getLattice<T,DESCRIPTOR>("fluid1"));
        particlesFluidAndWss.push_back(particles.get());
        particlesFluidAndWss.push_back(temporalMeanWss_plot);
        particlesFluidAndWss.push_back(&fsFields.getScalar<int>("flag"));
        particlesFluidAndWss.push_back(&fsFields.getScalar<T>("mass"));
        particlesFluidAndWss.push_back(&fsFields.getScalar<T>("volumeFraction"));
        particlesFluidAndWss.push_back(erosionDepositionPlot);


        multiBlockForUnknownAD.push_back(particles.get());
        multiBlockForUnknownAD.push_back(&fsFields.getLattice<T,DESCRIPTOR>("fluid1"));
        multiBlockForUnknownAD.push_back(&fsFields.getScalar<int>("flag"));

        multiBlockForTemporalMeanWss.push_back(&fsFields.getLattice<T,DESCRIPTOR>("fluid1"));
        multiBlockForTemporalMeanWss.push_back(wss);
        multiBlockForTemporalMeanWss.push_back(temporalMeanWss);
        multiBlockForTemporalMeanWss.push_back(&fsFields.getScalar<int>("flag"));
    }


    // std::string someNews = global::directories().getImageOutDir() + "energyEvol_inlet200m.dat";
    string ofilePath = outputDir + "energyEvol_inlet200m.dat";
    plb_ofstream ofile( ofilePath.c_str() );


    // Array<plint,3> iniDomCoord0 = getStlCoordFromReal(0.5, 0.009, 0, lbox);
    // Array<plint,3> iniDomCoord1 = getStlCoordFromReal(0.97, 0.1165, 0, lbox);
    // Box3D iniDomain(iniDomCoord0[0], iniDomCoord1[0], iniDomCoord0[1], iniDomCoord1[1], lbox.z0, lbox.z1);
    // applyProcessingFunctional (new RemoveMass3D<T,DESCRIPTOR>, iniDomain, fsSetup.freeSurfaceArgs);


    writeInterfaceSTL( fsFields.getScalar<T>("volumeFraction"), 99999);

    pcout << "Use particles : " << useParticles << std::endl;


    plint i = 0;
    T lastIterationTime = T();
    T totalSimTime = 0.;
    plint maxIter = (plint) simTime / dt;

    // T initialMass = freeSurfaceTotalMass<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, fsFields.getBoundingBox());

    pcout << "Entering main loop" << std::endl;
    while ( /*!velocityTracer.hasConverged() &&*/ i < maxIter )
    {
	    global::timer("iteration").restart();
        //pcout << "Iteration " << i << std::endl;


        // applyProcessingFunctional(new PouringLiquid3D<T,DESCRIPTOR>(dynamics , momentumAtInlet), // CF multiPhysics/freesurfaceboundarycondition for Pouring
        //                               inletDomain,
        //                               fsSetup.freeSurfaceArgs);

        fsFields.get("fluid1").executeInternalProcessors();

        if(i >= beginParticleDynamics )
        {
            if (useParticles)
            {
                // refresh dynamics : NoDynamics for cells that arent fluid or interface, else dynamicsParticle.
                applyProcessingFunctional(new ChangeDynamicForAdvectionDiffusionDynamic3D<T,DESCRIPTOR,T,DESCRIPTOR2>(omegaAD,dynamicsParticles),
                                          eliminateParticles,
                                          multiBlockChangeADdynamics/*,6*/);

                if (i% callDepositionErosion == 0) //critical number --> 100*criticalTime
                {
                    pcout << "Recalculating the distance" << endl;
                    global::timer("distance").restart();
                    distanceFunction = computeDistanceFunction(fsFields.getLattice<T,DESCRIPTOR>("fluid1"), fsFields.getLattice<T,DESCRIPTOR>("fluid1").getNy(), (T)1.0e-4).release();
                    pcout << "Time spent for recalculating the distance : " << global::timer("distance").stop() << endl;
                }

                computeWallShearStress(fsFields.getLattice<T,DESCRIPTOR>("fluid1"), *computeWallUnitNormal(*distanceFunction, fsFields.getBoundingBox()), *wss, wss->getBoundingBox() );

                // average WSS over neighbors, storing the results into temporalMeanWss.
                applyProcessingFunctional(new CalculateTemporalWssAverage3D<T,DESCRIPTOR>(), fsFields.getBoundingBox(), multiBlockForTemporalMeanWss );

                if (i% callDepositionErosion == 0) //critical number --> 100*criticalTime
                {
                    //temporalMeanWss_plot = temporalMeanWss * 1. / callDepositionErosion
                    multiply(T(1)/T(callDepositionErosion), *temporalMeanWss, *temporalMeanWss_plot, temporalMeanWss->getBoundingBox());
                    setToConstant(*temporalMeanWss, temporalMeanWss->getBoundingBox(), T(0));

                    // transform certain nextToWall fluids cells into erodableWall (DEPOSITION)
                    applyProcessingFunctional(new MakeThrombosisFromWssFunctional3D<T,DESCRIPTOR,T,DESCRIPTOR2,T>(stressThreshold,
                                                                                                                  criticalTime,
                                                                                                                  criticalNumber,
                                                                                                                  callDepositionErosion),
                                             depositionErosionZone,
                                             particlesFluidAndWss);

                    // transform certain erodableWall cells into fluid (bulk nature) cells (EROSION)
                    applyProcessingFunctional(new ErosionFromWssFunctional3D<T,DESCRIPTOR,T,DESCRIPTOR2,T>(stressThreshold,
                                                                                                           criticalTime,
                                                                                                           criticalNumber,
                                                                                                           callDepositionErosion,
                                                                                                           omega,
                                                                                                           cSmago,
                                                                                                           externalForce,
                                                                                                           dynamics,
                                                                                                           dynamicsParticles),
                                             depositionErosionZone,
                                             particlesFluidAndWss);

                    // refresh nextToWall/bulk node natures.
                    applyProcessingFunctional(new IniWallFlagRiverTestFunctional3D<T,DESCRIPTOR>(),
                                              fsFields.getBoundingBox(),
                                              fsFields.getLattice<T,DESCRIPTOR>("fluid1"), fsFields.getScalar<int>("flag"));
                    // Refresh dynamics:
                    // Initialization of AD dyamics inner border (for all erodable or not erodable walls that have neighbour nextToWall or bulk, set dynamicsParticle, else NoDynamics).
                    applyProcessingFunctional(new MakeThrombosisToParticlesFunctional3D<T,DESCRIPTOR,T,DESCRIPTOR2>(omegaAD,criticalNumber,dynamicsParticles),
                                                depositionErosionZone, fsFields.getLattice<T,DESCRIPTOR>("fluid1"), *particles);


                }
                applyProcessingFunctional(new CopyVelocityToParticlesFunctional3D<T,DESCRIPTOR,T,DESCRIPTOR2>(settlingVelocityVector),
                                          fsFields.getBoundingBox(), fsFields.getLattice<T,DESCRIPTOR>("fluid1"), *particles);
    		}

            // pcout << "a" << std::endl;
    	    applyProcessingFunctional ( new ComputeSmagoViscosityFunctional3D<T,DESCRIPTOR>(omega,cSmago),
    				                    fsFields.getBoundingBox(),
    					                fsFields.getLattice<T,DESCRIPTOR>("fluid1"));

            // pcout << "b" << std::endl;
            if (useParticles)
            {
                // for each particle cell, if dynamics is BGKSmago, transform it into AdvectionDiffusion.
                applyProcessingFunctional ( new TestTurbulentAdvectionDiffusionFunctional3D<T,DESCRIPTOR,T,DESCRIPTOR2>(),
                                            fsFields.getBoundingBox(),
                                            fsFields.getLattice<T,DESCRIPTOR>("fluid1"), *particles );

                particles->collideAndStream();

                // at each place where particles dynamics is BounceBack or NoDynamics or lattice is interface, each population pointing to fluid takes the value of this fluid cells population.
                // This is (probably?) done in order to ensure the stability. This overhides the bounce back, and it is like a zero-gradient condition.
                applyProcessingFunctional(new CopyUnknownFromNeighsFunctional3D<T,DESCRIPTOR2,T,DESCRIPTOR>(),
                                          particles->getBoundingBox(), multiBlockForUnknownAD);
            }

    	}

        // pcout << "cccc" << std::endl;
        // rewrite if condition in terms of volume considerations.
        if(i % removeMassIter == 0 && i > 0 /*&& i >= fluidAtRest*/) /* de base : i % 2000 sans fluid at rest */
        {
            // pcout << "	REMOVING MASS " << endl;
            // T averageMass = freeSurfaceAverageMass<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, outletDomain);
            // pcout << "Average Mass before: " << averageMass  << endl;
            // T averageVolumeFraction = freeSurfaceAverageVolumeFraction<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, outletDomain);
            // pcout << "Average Volume-Fraction before: " << setprecision(12) << averageVolumeFraction  << endl;
            // // rho = 1, u = 0, m = 0, vof = 0, flag stays interface


            T totalMass = freeSurfaceTotalMass<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, fsFields.getBoundingBox());
            pcout << "  Total mass before: " << totalMass  << endl;

            // T deltaMass = totalMass - initialMass;
            // pcout << "  Delta mass: " << deltaMass << endl;
            // if(deltaMass > (T)0){
            //     plint nx = outletDomain.getNx();
            //     // plint ny = outletDomain.getNy();
            //     // plint nz = outletDomain.getNz();
            //     if((plint)deltaMass < nx){
            //         nx -= (plint)deltaMass;
            //     }
            //     Box3D adaptedOutlet;
            //     adaptedOutlet.x0 = outletDomain.x0;
            //     adaptedOutlet.x1 = outletDomain.x0 + nx;
            //     adaptedOutlet.y0 = outletDomain.y0;
            //     adaptedOutlet.y1 = outletDomain.y1;
            //     adaptedOutlet.z0 = outletDomain.z0;
            //     adaptedOutlet.z1 = outletDomain.z1;
            //     applyProcessingFunctional (new RemoveMass3D<T,DESCRIPTOR>, outletDomain, fsSetup.freeSurfaceArgs);
            // }

            applyProcessingFunctional (new RemoveMass3D<T,DESCRIPTOR>, outletDomain, fsSetup.freeSurfaceArgs);

            totalMass = freeSurfaceTotalMass<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, fsFields.getBoundingBox());
            pcout << "  Total mass after: " << totalMass  << endl;


            // averageMass = freeSurfaceAverageMass<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, outletDomain);
            // pcout << "Average Mass after: " << averageMass  << endl;
            // averageVolumeFraction = freeSurfaceAverageVolumeFraction<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, outletDomain);
            // pcout << "Average Volume-Fraction after: " << setprecision(12) << averageVolumeFraction  << endl;
        }
        // pcout << "d" << std::endl;

        if (i%minStatIter == 0)
        {
            pcout <<"iter" << i <<"; T= " << currentTime << "; " << std::endl ;
            T latticeAverageEnergy 	= computeAverageEnergy(fsFields.getLattice<T,DESCRIPTOR>("fluid1"))*util::sqr(dx/dt);
            pcout  << "Average energy calculated with lattice: " << latticeAverageEnergy << std::endl;
            T averageMass = freeSurfaceAverageMass<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, fsFields.getBoundingBox());
            pcout << "Average Mass: " << averageMass  << endl;
            T averageDensity = freeSurfaceAverageDensity<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, fsFields.getBoundingBox());
            pcout << "Average Density: " << setprecision(12) << averageDensity  << endl;
            T averageVolumeFraction = freeSurfaceAverageVolumeFraction<T,DESCRIPTOR>(fsSetup.freeSurfaceArgs, fsFields.getBoundingBox());
            pcout << "Average Volume-Fraction: " << setprecision(12) << averageVolumeFraction  << endl;

            /*
            if(useParticles)
            {
                pcout << "; av particles=" << setprecision(10) << getStoredAverageDensity<T>(*particles) << endl;
            }*/
            ofile  << i << " " 	<< latticeAverageEnergy   	<< " "
                    << averageMass 	<< " "
                    << averageVolumeFraction <<   endl;

            totalSimTime += lastIterationTime;
            pcout << "Time of last iteration = " << lastIterationTime << " seconds" << endl;
            //pcout << "Total simulation time = " << totalSimTime/3600. << " hours" << endl;
            //pcout << "Estimated remaining time = " << (maxIter - i)*lastIterationTime/3600. << " hours" << endl;
        }

        if ( i%dataWrite==0) //&& i>= beginParticleDynamics+1 )
        {
            pcout << "*** Writing data ***" << endl;
            global::timer("images").restart();

            writeInterfaceSTL(fsFields.getScalar<T>("volumeFraction"), i);

            applyProcessingFunctional ( new CopyScalarFromExternalFieldToScalarField<T,DESCRIPTOR>(),
                                       fsFields.getBoundingBox(), fsFields.getLattice<T,DESCRIPTOR>("fluid1"), flagMatrixPlot );

            // Box3D myBox = fsFields.getBoundingBox().enlarge(-1);
            // myBox.y0 = myBox.y1 - myBox.getNy()/2;
            // myBox.y1 = myBox.y1 - myBox.getNy()/8;
            // writeUVtk(fsFields.getLattice<T,DESCRIPTOR>("fluid1"), myBox, "myUvtk", location, dx, dt, i);

            // save bounceback domain:
            /*
            {
                Box3D innerDomain(fsFields.getBoundingBox());
                MultiScalarField3D<T> boolMask(innerDomain.getNx(), innerDomain.getNy(), innerDomain.getNz());
                plint fluidFlag = 1;
                plint wallFlag = -10;
                applyProcessingFunctional(new BoolMaskFromBounceBack3D<T,DESCRIPTOR,T>(fluidFlag, wallFlag),
                                          fsFields.getBoundingBox(), fsFields.getLattice<T,DESCRIPTOR>("fluid1"), boolMask);

                std::vector<T> isoLevels;
                isoLevels.push_back((T) 0.5);
                typedef TriangleSet<T>::Triangle Triangle;
                std::vector<Triangle> triangles;
                isoSurfaceMarchingCube(triangles, boolMask, isoLevels, boolMask.getBoundingBox());
                TriangleSet<T>(triangles).writeBinarySTL(myPath + createFileName("converter", i, 6) + ".stl");
            }
            */
            if (useParticles)
            {
                applyProcessingFunctional ( new CopyStateOfLocalDynamicToScalarField<T,DESCRIPTOR2>(),
                                             particles->getBoundingBox(), *particles, *particlesDynamicsPlot );
                //now particlesDynamicsPlot contains 1 where noDynamics, 0 else.

                writeVTKwholeLattice(fsFields.getLattice<T,DESCRIPTOR>("fluid1"),
                                     particles.get(),
                                     fsFields.getScalar<T>("volumeFraction"),
                                     fsFields.getScalar<int>("flag"),
                                     flagMatrixPlot,
                                     particlesDynamicsPlot,
                                     erosionDepositionPlot,
                                     distanceFunction,
                                     wss,
                                     temporalMeanWss_plot,
                                     location,
                                     dx,
                                     dt,
                                     i);
            }
            pcout << "Time spent for writing big VTK : " << global::timer("images").stop() << endl;

        } //end if write data

        //  if (i%100 == 0)
            //  pcout << "------------------------------- " << i << " ---------------------------" << std::endl << std::endl;

    //         if (i%convergenceIter==0) {
    //             velocityTracer.takeValue(computeAverageEnergy(fsFields.getLattice<T,DESCRIPTOR>("fluid1")));
    //         }
    //
        ++i;
        currentTime = i*dt;
        lastIterationTime = global::timer("iteration").stop();
    }

    pcout << " end of simulation--> either converged(diverged) or end of time steps " << std::endl;

}
