/* This file is part of the Palabos library.
 * Copyright (C) 2009 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of Palabos can be downloaded at 
 * <http://www.lbmethod.org/palabos/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef NORMAL_AND_TANGENT_H
#define NORMAL_AND_TANGENT_H

#include "core/globalDefs.h"
#include "atomicBlock/blockLattice3D.h"
#include "poissonEquation3D.h"
#include <memory>

namespace plb {

template<typename T>
class CheckScalarFunctional3D : public BoxProcessingFunctional3D_S<T>
{
public:
    virtual void process(Box3D domain, ScalarField3D<T>& data) {
        for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
            for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
                for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                    if (data.get(iX,iY,iZ)<-1.e100) {
                        pcout << "is smaller" << std::endl;
                    }
                }
            }
        }
    }
    virtual CheckScalarFunctional3D<T>* clone() const {
        return new CheckScalarFunctional3D<T>(*this);
    }
    virtual void getModificationPattern(std::vector<bool>& isWritten) const {
        isWritten[0] = false;
    }
    virtual BlockDomain::DomainT appliesTo() const {
        return BlockDomain::bulkAndEnvelope;
    }
};

template<typename T>
class CheckTensorFunctional3D : public BoxProcessingFunctional3D_T<T,3>
{
public:
    virtual void process(Box3D domain, TensorField3D<T,3>& data) {
        for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
            for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
                for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                    if (data.get(iX,iY,iZ)[0]<-1.e100) {
                        pcout << "is smaller" << std::endl;
                    }
                    if (data.get(iX,iY,iZ)[1]<-1.e100) {
                        pcout << "is smaller" << std::endl;
                    }
                    if (data.get(iX,iY,iZ)[2]<-1.e100) {
                        pcout << "is smaller" << std::endl;
                    }
                }
            }
        }
    }
    virtual CheckTensorFunctional3D<T>* clone() const {
        return new CheckTensorFunctional3D<T>(*this);
        }
    virtual void getModificationPattern(std::vector<bool>& isWritten) const {
        isWritten[0] = false;
    }
    virtual BlockDomain::DomainT appliesTo() const {
        return BlockDomain::bulkAndEnvelope;
    }
};


template<typename T>
class ComputeNormalGradientFunctional3D : public BoxProcessingFunctional3D_ST<T,T,3>
{
public:
    ComputeNormalGradientFunctional3D(Box3D outerBound_);
    virtual void process(Box3D domain, ScalarField3D<T>& data, TensorField3D<T,3>& gradient);
    virtual ComputeNormalGradientFunctional3D<T>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    static bool isWall(T value);
    Array<T,3> computeGradient(ScalarField3D<T>& data, Dot3D pos, plint iX, plint iY, plint iZ) const;
    T xp1(ScalarField3D<T> const& data, Dot3D pos, plint iX, plint iY, plint iZ ) const;
    T xm1(ScalarField3D<T> const& data, Dot3D pos, plint iX, plint iY, plint iZ ) const;
    T yp1(ScalarField3D<T> const& data, Dot3D pos, plint iX, plint iY, plint iZ ) const;
    T ym1(ScalarField3D<T> const& data, Dot3D pos, plint iX, plint iY, plint iZ ) const;
    T zp1(ScalarField3D<T> const& data, Dot3D pos, plint iX, plint iY, plint iZ ) const;
    T zm1(ScalarField3D<T> const& data, Dot3D pos, plint iX, plint iY, plint iZ ) const;
private:
    Box3D outerBound;
};

template<typename T>
ComputeNormalGradientFunctional3D<T>::ComputeNormalGradientFunctional3D(Box3D outerBound_)
    : outerBound(outerBound_)
{ }

template<typename T>
bool ComputeNormalGradientFunctional3D<T>::isWall(T value) {
    static const T epsilon=1.e-10;
    if (value<epsilon) {
        return true;
    }
    else {
        return false;
    }
}

template<typename T>
T ComputeNormalGradientFunctional3D<T>::xm1( ScalarField3D<T> const& data,
                                             Dot3D pos, plint iX, plint iY, plint iZ) const
{
    // First, make sure I am not on an outer bound, because in that case,
    //   accessing the next neighbor would create an access violation.
    //   Then, if the next neighbor is a wall node, assume that the wall
    //   is half-way between nodes, and impose a zero condition.
    if (iX-1+pos.x < outerBound.x0 || isWall(data.get(iX-1,iY,iZ))) {
        return -data.get(iX,iY,iZ);
    }
    else {
        return data.get(iX-1,iY,iZ);
    }
}

template<typename T>
T ComputeNormalGradientFunctional3D<T>::xp1( ScalarField3D<T> const& data,
                                             Dot3D pos, plint iX, plint iY, plint iZ) const
{
    // First, make sure I am not on an outer bound, because in that case,
    //   accessing the next neighbor would create an access violation.
    //   Then, if the next neighbor is a wall node, assume that the wall
    //   is half-way between nodes, and impose a zero condition.
    if (iX+1+pos.x > outerBound.x1 || isWall(data.get(iX+1,iY,iZ))) {
        return -data.get(iX,iY,iZ);
    }
    else {
        return data.get(iX+1,iY,iZ);
    }
}

template<typename T>
T ComputeNormalGradientFunctional3D<T>::ym1( ScalarField3D<T> const& data,
                                             Dot3D pos, plint iX, plint iY, plint iZ) const
{
    // First, make sure I am not on an outer bound, because in that case,
    //   accessing the next neighbor would create an access violation.
    //   Then, if the next neighbor is a wall node, assume that the wall
    //   is half-way between nodes, and impose a zero condition.
    if (iY-1+pos.y < outerBound.y0 || isWall(data.get(iX,iY-1,iZ))) {
        return -data.get(iX,iY,iZ);
    }
    else {
        return data.get(iX,iY-1,iZ);
    }
}

template<typename T>
T ComputeNormalGradientFunctional3D<T>::yp1( ScalarField3D<T> const& data,
                                             Dot3D pos, plint iX, plint iY, plint iZ) const
{
    // First, make sure I am not on an outer bound, because in that case,
    //   accessing the next neighbor would create an access violation.
    //   Then, if the next neighbor is a wall node, assume that the wall
    //   is half-way between nodes, and impose a zero condition.
    if (iY+1+pos.y > outerBound.y1 || isWall(data.get(iX,iY+1,iZ))) {
        return -data.get(iX,iY,iZ);
    }
    else {
        return data.get(iX,iY+1,iZ);
    }
}

template<typename T>
T ComputeNormalGradientFunctional3D<T>::zm1( ScalarField3D<T> const& data,
                                             Dot3D pos, plint iX, plint iY, plint iZ) const
{
    // First, make sure I am not on an outer bound, because in that case,
    //   accessing the next neighbor would create an access violation.
    //   Then, if the next neighbor is a wall node, assume that the wall
    //   is half-way between nodes, and impose a zero condition.
    if (iZ-1+pos.z < outerBound.z0 || isWall(data.get(iX,iY,iZ-1))) {
        return -data.get(iX,iY,iZ);
    }
    else {
        return data.get(iX,iY,iZ-1);
    }
}

template<typename T>
T ComputeNormalGradientFunctional3D<T>::zp1( ScalarField3D<T> const& data,
                                             Dot3D pos, plint iX, plint iY, plint iZ) const
{
    // First, make sure I am not on an outer bound, because in that case,
    //   accessing the next neighbor would create an access violation.
    //   Then, if the next neighbor is a wall node, assume that the wall
    //   is half-way between nodes, and impose a zero condition.
    if (iZ+1+pos.z > outerBound.z1 || isWall(data.get(iX,iY,iZ+1))) {
        return -data.get(iX,iY,iZ);
    }
    else {
        return data.get(iX,iY,iZ+1);
    }
}

template<typename T>
Array<T,3> ComputeNormalGradientFunctional3D<T>::computeGradient (
        ScalarField3D<T>& data, Dot3D pos, plint iX, plint iY, plint iZ ) const
{
    PLB_PRECONDITION(iX>0 && iX<data.getNx()-1);
    PLB_PRECONDITION(iY>0 && iY<data.getNy()-1);
    PLB_PRECONDITION(iZ>0 && iZ<data.getNz()-1);

    Array<T,3> gradient;
    gradient[0] = 0.5*(-xp1(data,pos,iX,iY,iZ)+xm1(data,pos,iX,iY,iZ));
    gradient[1] = 0.5*(-yp1(data,pos,iX,iY,iZ)+ym1(data,pos,iX,iY,iZ));
    gradient[2] = 0.5*(-zp1(data,pos,iX,iY,iZ)+zm1(data,pos,iX,iY,iZ));
    return gradient;
}

template<typename T>
void ComputeNormalGradientFunctional3D<T>::process (
        Box3D domain, ScalarField3D<T>& data, TensorField3D<T,3>& gradient )
{
    Dot3D offset = computeRelativeDisplacement(data, gradient);
    Dot3D pos = data.getLocation();
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                T& value = data.get(iX,iY,iZ);
                Array<T,3>& grad = gradient.get(iX+offset.x, iY+offset.y, iZ+offset.z);
                if (isWall(value)) {
                    grad.resetToZero();
                }
                else {
                    grad = computeGradient(data, pos, iX, iY, iZ);
                    T norm = sqrt(VectorTemplateImpl<T,3>::normSqr(grad));
                    if (norm>1.e-10) {
                        grad /= norm;
                    }
                    else {
                        grad.resetToZero();
                    }
                }
            }
        }
    }
}

template<typename T>
ComputeNormalGradientFunctional3D<T>* ComputeNormalGradientFunctional3D<T>::clone() const
{
    return new ComputeNormalGradientFunctional3D<T>(*this);
}

template<typename T>
void ComputeNormalGradientFunctional3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    modified[1] = modif::dataStructure;
}

/// Two ortho-normal vectors, orthogonal to a given normal 3D vector, using Gram-Schmidt orthogonalization.
template<typename T>
void tangentsFromNormal(Array<T,3> const& unitNormal, Array<T,3>& tangent1, Array<T,3>& tangent2) {
    // Determine the two smallest components of unitNormal. To avoid potential singularities,
    //   the Gram-Schmidt procedure will be applied to the Euclidean unit vectors pointing in
    //   direction of these two smallest components.
    Array<int,3> indices(0,1,2);
    Array<T,3> sortedNormal(unitNormal);
    if (sortedNormal[2] < sortedNormal[1]) {
        std::swap(sortedNormal[1], sortedNormal[2]);
        std::swap(indices[1], indices[2]);
    }
    if (sortedNormal[1] < sortedNormal[0]) {
        std::swap(sortedNormal[0], sortedNormal[1]);
        std::swap(indices[0], indices[1]);
    }
    if (sortedNormal[2] < sortedNormal[1]) {
        std::swap(sortedNormal[1], sortedNormal[2]);
        std::swap(indices[1], indices[2]);
    }

    // The first tangential vector is equal to the direction of the smallest component of
    //   unitNormal, minus projection onto unitNormal.
    tangent1.resetToZero();
    tangent1[indices[0]] = (T)1;

    tangent1 -= VectorTemplateImpl<T,3>::scalarProduct(tangent1, unitNormal);
    tangent1 /= sqrt(VectorTemplateImpl<T,3>::normSqr(tangent1));
    
    // The second tangential vector is equal to the direction of the second-to-smallest component
    //   of unitNormal, minus projection onto unitNormal, minus projection onto tangent1.
    tangent2.resetToZero();
    tangent2[indices[1]] = (T)1;

    tangent2 -= VectorTemplateImpl<T,3>::scalarProduct(tangent2, unitNormal) -
                VectorTemplateImpl<T,3>::scalarProduct(tangent2, tangent1);
    tangent2 /= sqrt(VectorTemplateImpl<T,3>::normSqr(tangent2));
}

template<typename T>
class ComputeTangentFromNormal3D : public BoxProcessingFunctional3D_TT<T,3,T,3>
{
public:
    ComputeTangentFromNormal3D(int whichTangent_);
    virtual void process(Box3D domain, TensorField3D<T,3>& normal, TensorField3D<T,3>& tangent);
    virtual ComputeTangentFromNormal3D<T>* clone() const;
    virtual void getModificationPattern(std::vector<bool>& isWritten) const;
    virtual BlockDomain::DomainT appliesTo() const;
private:
    int whichTangent;
};


template<typename T>
ComputeTangentFromNormal3D<T>::ComputeTangentFromNormal3D(int whichTangent_)
    : whichTangent(whichTangent_)
{ }

template<typename T>
void ComputeTangentFromNormal3D<T>::process (
        Box3D domain, TensorField3D<T,3>& normal, TensorField3D<T,3>& tangent )
{
    Array<T,3> dummyTangent;
    Dot3D offset = computeRelativeDisplacement(normal, tangent);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                if (VectorTemplateImpl<T,3>::normSqr(normal.get(iX,iY,iZ))<1.e-10) {
                    tangent.get(iX,iY,iZ).resetToZero();
                }
                else {
                    if (whichTangent==0) {
                        tangentsFromNormal(normal.get(iX,iY,iZ),
                                           tangent.get(iX+offset.x,iY+offset.y,iZ+offset.z), dummyTangent);
                    }
                    else {
                        tangentsFromNormal(normal.get(iX,iY,iZ),
                                           dummyTangent, tangent.get(iX+offset.x,iY+offset.y,iZ+offset.z));
                    }
                }
            }
        }
    }
}

template<typename T>
ComputeTangentFromNormal3D<T>* ComputeTangentFromNormal3D<T>::clone() const
{
    return new ComputeTangentFromNormal3D<T>(*this);
}

template<typename T>
void ComputeTangentFromNormal3D<T>::getModificationPattern(std::vector<bool>& isWritten) const {
    isWritten[0] = false;
    isWritten[1] = true;
}

template<typename T>
BlockDomain::DomainT ComputeTangentFromNormal3D<T>::appliesTo() const {
    return BlockDomain::bulk;
}



template<typename T, template<class U> class Descriptor>
class ComputeShearStressFunctional3D : public BoxProcessingFunctional3D
{
public:
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual ComputeShearStressFunctional3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};


template<typename T, template<class U> class Descriptor>
void ComputeShearStressFunctional3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks )
{
    PLB_PRECONDITION( atomicBlocks.size()==3 );
    BlockLattice3D<T,Descriptor>& lattice = dynamic_cast<BlockLattice3D<T,Descriptor>&>(*atomicBlocks[0]);
    TensorField3D<T,3>& normal = dynamic_cast<TensorField3D<T,3>&>(*atomicBlocks[1]);
    ScalarField3D<T>& shearStress = dynamic_cast<ScalarField3D<T>&>(*atomicBlocks[2]);

    Dot3D offsetNormal = computeRelativeDisplacement(lattice, normal);
    Dot3D offsetStress = computeRelativeDisplacement(lattice, shearStress);

    Array<T,SymmetricTensor<T,Descriptor>::n> S;
    Array<T,3> wallStress, tangentialStress;

    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                Cell<T,Descriptor> const& cell = lattice.get(iX,iY,iZ);
                cell.computeShearStress(S);
                T omega     = cell.getDynamics().getOmega();
                T rhoBar    = cell.getDynamics().computeRhoBar(cell);
                
                T prefactor = - omega * Descriptor<T>::invCs2 * Descriptor<T>::invRho(rhoBar) / (T)2;
                for (int iTensor=0; iTensor<SymmetricTensor<T,Descriptor>::n; ++iTensor) {
                    S[iTensor] *= prefactor;
                }
                
                Array<T,3> const& normalVect = normal.get(iX+offsetNormal.x, iY+offsetNormal.y, iZ+offsetNormal.z);
                SymmetricTensor<T,Descriptor>::matVectMult(S, normalVect, wallStress);
                tangentialStress = wallStress -
                                   normalVect*VectorTemplateImpl<T,3>::scalarProduct(wallStress, normalVect);
                shearStress.get(iX+offsetStress.x, iY+offsetStress.y, iZ+offsetStress.z) 
                    = sqrt(VectorTemplateImpl<T,3>::normSqr(tangentialStress));
            }
        }
    }
}

template<typename T, template<class U> class Descriptor>
ComputeShearStressFunctional3D<T,Descriptor>* ComputeShearStressFunctional3D<T,Descriptor>::clone() const
{
    return new ComputeShearStressFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<class U> class Descriptor>
void ComputeShearStressFunctional3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    modified[1] = modif::nothing;
    modified[2] = modif::staticVariables;
}


template<typename T, template<class U> class Descriptor>
void computeWallShearStress ( MultiBlockLattice3D<T,Descriptor>& lattice,
                              MultiTensorField3D<T,3>& wallUnitNormal,
                              MultiScalarField3D<T>& wss,
                              Box3D domain )
{
    //applyProcessingFunctional(new CheckTensorFunctional3D<T>, domain, wallUnitNormal);
    std::vector<MultiBlock3D*> multiBlocks;
    multiBlocks.push_back(&lattice);
    multiBlocks.push_back(&wallUnitNormal);
    multiBlocks.push_back(&wss);
    applyProcessingFunctional(new ComputeShearStressFunctional3D<T,Descriptor>, domain, multiBlocks);
    //applyProcessingFunctional(new CheckScalarFunctional3D<T>, domain, wss);
}

template<typename T, template<class U> class Descriptor>
std::auto_ptr<MultiScalarField3D<T> > computeWallShearStress (
                                    MultiBlockLattice3D<T,Descriptor>& lattice,
                                    MultiTensorField3D<T,3>& wallUnitNormal,
                                    Box3D domain )
{
    MultiScalarField3D<T>* wss = new MultiScalarField3D<T>(lattice, domain);
    computeWallShearStress(lattice, wallUnitNormal, *wss, domain);
    return std::auto_ptr<MultiScalarField3D<T> >(wss);
}
        
template<typename T, template<class U> class Descriptor>
std::auto_ptr<MultiScalarField3D<T> > computeWallShearStress (
                                    MultiBlockLattice3D<T,Descriptor>& lattice,
                                    MultiTensorField3D<T,3>& wallUnitNormal )
{
    return computeWallShearStress(lattice, wallUnitNormal, lattice.getBoundingBox());
}

template<typename T, template<class U> class Descriptor>
void computeDistanceFunction(MultiBlockLattice3D<T,Descriptor>& lattice,
                             MultiScalarField3D<T>& distanceFunction,
                             Box3D domain,
                             plint Nx,
                             T accuracy)
{
    // Given that the Poisson operator accesses nearest-neighbors, it cannot be applied to
    //   boundary cells.
    Box3D innerDomain(domain.enlarge(-1));
    MultiScalarField3D<T> boolMask(lattice, domain);
    plint fluidFlag = 1;
    plint wallFlag = -10;
    applyProcessingFunctional(new BoolMaskFromBounceBack3D<T,Descriptor,T>(fluidFlag, wallFlag),
                              domain, lattice, boolMask);
    // now boolMask contains wallFlag where cells dynamics is bounceBack, fluidFlag else.
    // Say that all outer cells are walls, to make sure the solver converges.
    plint nx = boolMask.getNx();
    plint ny = boolMask.getNy();
    plint nz = boolMask.getNz();
    setToConstant(boolMask, Box3D(0,    nx-1, 0,    ny-1, 0,    0   ), (T)-10.);
    setToConstant(boolMask, Box3D(0,    nx-1, 0,    ny-1, nz-1, nz-1), (T)-10.);
    setToConstant(boolMask, Box3D(0,    nx-1, 0,    0,    0,    nz-1), (T)-10.);
    setToConstant(boolMask, Box3D(0,    nx-1, ny-1, ny-1, 0,    nz-1), (T)-10.);
    setToConstant(boolMask, Box3D(0,    0,    0,    ny-1, 0,    nz-1), (T)-10.);
    setToConstant(boolMask, Box3D(nx-1, nx-1, 0,    ny-1, 0,    nz-1), (T)-10.);

    applyProcessingFunctional(new DetectBoundaryFunctional3D<T>, innerDomain, boolMask);
    //now all fluidFlag cells with wall neighbor have the value 0. boolMask is not a boolean (can take three values : 1, 0 and -10. 

    setToConstant(distanceFunction, domain, T());
    T rhs = -8./Nx/Nx; // ?????
    T beta = 1.;
    plint i=0;
    T difference=1.;
    
    util::ValueTracer<T> converge((T)1, Nx, 1.0e-3);
    while(difference>accuracy)
    {
        DirichletPoissonFunctional3D<T> functional(rhs, beta);
        applyProcessingFunctional(functional, innerDomain, distanceFunction, boolMask);
        // now distanceFunction is filled according to boolMask's walls, omega(=beta) and rhs
        difference = functional.getMaximumDifference();
        converge.takeValue(difference);
        if (i++%1==100) {// ????
            pcout << "difference=" << difference << std::endl;
        }
        if (converge.hasConverged()) break;
    }
    // normalization (why??)
    pcout << "max dist function = " << computeMax(distanceFunction) << std::endl;
    divideInPlace(distanceFunction, computeMax(distanceFunction), distanceFunction.getBoundingBox());

    //applyProcessingFunctional(new CheckScalarFunctional3D<T>, domain, distanceFunction);
    //     distanceFunction(x,y,z) = 1. - sqrt(1. - |distanceFunction(x,y,z)|):
    applyProcessingFunctional(new PoissonToDistanceFunctional3D<T>, domain, distanceFunction);
    //applyProcessingFunctional(new CheckScalarFunctional3D<T>, domain, distanceFunction);

}

template<typename T, template<class U> class Descriptor>
std::auto_ptr<MultiScalarField3D<T> > computeDistanceFunction (MultiBlockLattice3D<T,Descriptor>& lattice,
                                                               Box3D domain,
                                                               plint Nx,
                                                               T accuracy)
{
    MultiScalarField3D<T>* distanceFunction = new MultiScalarField3D<T>(lattice, domain);
    computeDistanceFunction(lattice, *distanceFunction, domain, Nx, accuracy);
    return std::auto_ptr<MultiScalarField3D<T> >(distanceFunction);
}

template<typename T, template<class U> class Descriptor>
std::auto_ptr<MultiScalarField3D<T> > computeDistanceFunction (MultiBlockLattice3D<T,Descriptor>& lattice,
                                                               plint Nx,
                                                               T accuracy)
{
    return computeDistanceFunction(lattice, lattice.getBoundingBox(), Nx, accuracy);
}
        
template<typename T>
void computeWallUnitNormal (
        MultiScalarField3D<T>& distanceFunction,
        MultiTensorField3D<T,3>& unitNormal, Box3D domain )
{
    //applyProcessingFunctional(new CheckScalarFunctional3D<T>, domain, distanceFunction);
    applyProcessingFunctional (
            // data = distanceFunction, gradient=unitNormal
            new ComputeNormalGradientFunctional3D<T>(domain), domain, distanceFunction, unitNormal );
    //applyProcessingFunctional(new CheckTensorFunctional3D<T>, domain, unitNormal);
}

template<typename T>
std::auto_ptr<MultiTensorField3D<T,3> > computeWallUnitNormal (
        MultiScalarField3D<T>& distanceFunction, Box3D domain )
{
    MultiTensorField3D<T,3>* unitNormal =
        new MultiTensorField3D<T,3>(distanceFunction, domain);
    computeWallUnitNormal(distanceFunction, *unitNormal, domain);
    return std::auto_ptr<MultiTensorField3D<T,3> > (unitNormal);
}

template<typename T>
std::auto_ptr<MultiTensorField3D<T,3> > computeWallUnitNormal(MultiScalarField3D<T>& distanceFunction)
{
    return computeWallUnitNormal(distanceFunction, distanceFunction.getBoundingBox());
}

}  // namespace plb

#endif  // NORMAL_AND_TANGENT_H
